#!/usr/bin/env rspec

# Copyright (c) [2018] SUSE LLC
#
# All Rights Reserved.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of version 2 of the GNU General Public License as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
# more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, contact SUSE LLC.
#
# To contact SUSE LLC about this file by physical or electronic mail, you may
# find current contact information at www.suse.com.

require_relative "../test_helper"
require "cwm/rspec"
require "y2partitioner/widgets/encrypt_password"

describe Y2Partitioner::Widgets::EncryptPassword do
  subject(:widget) { described_class.new(controller) }

  include_examples "CWM::AbstractWidget"

  let(:controller) { double("FilesystemController", password: "secret123") }
  let(:pw1) { "password1" }
  let(:pw2) { "password1" }
  let(:password_checker) { Y2Storage::EncryptPasswordChecker.new }
  let(:enabled) { true }

  before do
    allow(Y2Storage::EncryptPasswordChecker).to receive(:new).and_return(password_checker)
    allow(Yast::UI).to receive(:QueryWidget).with(Id(:pw1), :Value).and_return(pw1)
    allow(Yast::UI).to receive(:QueryWidget).with(Id(:pw2), :Value).and_return(pw2)

    # Validation always checks whether the widget is really active
    allow(Yast::UI).to receive(:QueryWidget).with(Id(widget.widget_id), :Enabled).and_return enabled
  end

  describe "#contents" do
    it "includes two password fields" do
      contents = widget.contents
      pw1 = contents.nested_find { |i| i.is_a?(Yast::Term) && i.params == [:pw1] }
      expect(pw1).to_not be_nil
      pw2 = contents.nested_find { |i| i.is_a?(Yast::Term) && i.params == [:pw2] }
      expect(pw2).to_not be_nil
    end

    it "prefills the password fields with the initial password" do
      contents = widget.contents
      field = contents.nested_find { |i| i.value == :Password }
      expect(field.params).to include "secret123"
    end
  end

  describe "#validate" do
    it "checkes the password" do
      expect(password_checker).to receive(:error_msg).with(pw1, pw2)
      widget.validate
    end

    context "when a good password is chosen" do
      before do
        allow(password_checker).to receive(:error_msg)
          .and_return(nil)
      end

      it "returns true" do
        expect(widget.validate).to eq(true)
      end
    end

    context "when a bad password is chosen" do
      before do
        allow(password_checker).to receive(:error_msg)
          .and_return("some error")
      end

      context "when the widget is active" do
        let(:enabled) { true }

        it "returns false" do
          allow(Yast::Report).to receive(:Error)
          expect(widget.validate).to eq(false)
        end

        it "displays an error message" do
          expect(Yast::Report).to receive(:Error).with("some error")
          widget.validate
        end
      end

      context "when the widget is disabled" do
        let(:enabled) { false }

        it "returns true" do
          expect(widget.validate).to eq(true)
        end
      end
    end
  end

  describe "#store" do
    it "assigns password to the controller" do
      expect(controller).to receive(:password=).with(pw1)
      widget.store
    end
  end

  describe "#cleanup" do
    it "cleans up the password checker" do
      expect(password_checker).to receive(:tear_down)
      widget.cleanup
    end
  end
end
